package deepinid

import (
	"fmt"
	"time"

	"pkg.deepin.io/daemon/sync/infrastructure/hardware"
	"pkg.deepin.io/daemon/sync/infrastructure/log"

	"github.com/fatih/structs"
	"github.com/godbus/dbus"
	"github.com/mitchellh/mapstructure"
)

const (
	dbusPropUserInfo    = "UserInfo"
	dbusPropMachineName = "MachineName"
)

// SetToken check token verify
func (d *DeepinID) SetToken(tokenInfo map[string]interface{}) *dbus.Error {
	log.Info("call", tokenInfo)

	token := fmt.Sprint(tokenInfo["token"])
	accessToken := fmt.Sprint(tokenInfo["access_token"])
	expiry, _ := time.Parse(time.RFC3339, fmt.Sprint(tokenInfo["expiry"]))

	method := d.GetDBusInterface() + ".SetToken"
	if len(token) == 0 || len(accessToken) == 0 {
		return dbus.NewError(method, []interface{}{"empty token"})
	}

	err := d.updateToken(token, accessToken, expiry.Unix())
	if err != nil {
		return dbus.NewError(method, []interface{}{err.Error()})
	}

	return nil
}

// GetHardware return hardware info
func (d *DeepinID) GetHardware() (*hardware.Hardware, *dbus.Error) {
	return d.hardware, nil
}

// GetToken return hardware info
func (d *DeepinID) GetToken() (string, *dbus.Error) {
	td := &tokenData{}
	mapstructure.Decode(d.userInfo, td)
	if td.isExpired() {
		//refresh
		log.Infof("token expired, get new: %v", d.userInfo)
		d.refreshToken(td.Token, td.AccessToken, td.Expiry)
		mapstructure.Decode(d.userInfo, td)
		return td.AccessToken, nil
	}
	return td.AccessToken, nil

}

func (d *DeepinID) isLogin() bool {
	td := d.getTokenData()
	return td.IsLoggedIn
}

// Login call dialog to login
func (d *DeepinID) Login() *dbus.Error {
	method := d.GetDBusInterface() + ".Login"

	if d.isLogin() {
		return dbus.NewError(method, []interface{}{fmt.Sprint("user has login")})
	}

	err := launchLoginApp()
	if err != nil {
		return dbus.NewError(method, []interface{}{err.Error()})
	}
	return nil
}

// Logout remove login info
func (d *DeepinID) Logout() *dbus.Error {
	method := d.GetDBusInterface() + ".Logout"

	if !d.isLogin() {
		return dbus.NewError(method, []interface{}{fmt.Sprint("user has NOT login")})
	}

	// remove token
	err := removeToken()
	if err != nil {
		return dbus.NewError(method, []interface{}{err.Error()})
	}
	userInfo := map[string]interface{}{}
	d.SetMapProp(&d.userInfo, dbusPropUserInfo, userInfo)
	return nil
}

// RevokeAccess remove login info
func (d *DeepinID) RevokeAccess(accessToken string) *dbus.Error {
	method := d.GetDBusInterface() + ".RevokeAccess"

	if !d.isLogin() {
		return dbus.NewError(method, []interface{}{fmt.Sprint("user has NOT login")})
	}

	td := d.getTokenData()
	if td.AccessToken != accessToken {
		return dbus.NewError(method, []interface{}{fmt.Sprint("user accessToken not match")})
	}

	td.Expiry = time.Now().Unix()

	m := structs.Map(td)
	d.SetMapProp(&d.userInfo, dbusPropUserInfo, m)

	td.AccessToken = ""
	saveToken(td)
	return nil
}

// func (d *DeepinID) Debug() *dbus.Error {
// 	td := d.getTokenData()
// 	userInfo := structs.Map(td)
// 	userInfo["tttttt"] = time.Now().Unix()
// 	d.SetMapProp(&d.userInfo, dbusPropUserInfo, userInfo)
// 	return nil

// }
