package hardware

import (
	"crypto/sha256"
	"encoding/json"
	"fmt"
	"io/ioutil"
)

type DMI struct {
	BiosDate       string `json:"bios_date"`
	BiosVendor     string `json:"bios_vendor"`
	BiosVersion    string `json:"bios_version"`
	BoardName      string `json:"board_name"`
	BoardVendor    string `json:"board_vendor"`
	BoardVersion   string `json:"board_version"`
	BoardSerial    string `json:"board_serial"`
	BoardAssetTag  string `json:"board_asset_tag"`
	ProductName    string `json:"product_name"`
	ProductFamily  string `json:"product_family"`
	ProductVersion string `json:"product_version"`
	ProductSerial  string `json:"product_serial"`
	ProductUUID    string `json:"product_uuid"`
}

const (
	sysfsDMIDir = "/sys/class/dmi/id"

	fileBiosDate       = "bios_date"
	fileBiosVendor     = "bios_vendor"
	fileBiosVersion    = "bios_version"
	fileBoardName      = "board_name"
	fileBoardVendor    = "board_vendor"
	fileBoardVersion   = "board_version"
	fileBoardSerial    = "board_serial"
	fileBoardAssetTag  = "board_asset_tag"
	fileProductName    = "product_name"
	fileProductFamily  = "product_family"
	fileProductVersion = "product_version"
	fileProductSerial  = "product_serial"
	fileProductUUID    = "product_uuid"
)

func GetDMI() (*DMI, error) {
	return doGetDMI(sysfsDMIDir)
}

func doGetDMI(dir string) (*DMI, error) {
	var d DMI
	var err error
	d.BiosDate, err = readFile(dir + "/" + fileBiosDate)
	if err != nil {
		return nil, err
	}
	d.BiosVendor, err = readFile(dir + "/" + fileBiosVendor)
	if err != nil {
		return nil, err
	}
	d.BiosVersion, err = readFile(dir + "/" + fileBiosVersion)
	if err != nil {
		return nil, err
	}
	d.BoardName, err = readFile(dir + "/" + fileBoardName)
	if err != nil {
		return nil, err
	}
	d.BoardVendor, err = readFile(dir + "/" + fileBoardVendor)
	if err != nil {
		return nil, err
	}
	d.BoardVersion, err = readFile(dir + "/" + fileBoardVersion)
	if err != nil {
		return nil, err
	}
	d.BoardSerial, err = readFile(dir + "/" + fileBoardSerial)
	if err != nil {
		return nil, err
	}
	d.BoardAssetTag, err = readFile(dir + "/" + fileBoardAssetTag)
	if err != nil {
		return nil, err
	}
	d.ProductName, err = readFile(dir + "/" + fileProductName)
	if err != nil {
		return nil, err
	}
	d.ProductFamily, err = readFile(dir + "/" + fileProductFamily)
	if err != nil {
		return nil, err
	}
	d.ProductVersion, err = readFile(dir + "/" + fileProductVersion)
	if err != nil {
		return nil, err
	}
	d.ProductSerial, err = readFile(dir + "/" + fileProductSerial)
	if err != nil {
		return nil, err
	}
	d.ProductUUID, err = readFile(dir + "/" + fileProductUUID)
	if err != nil {
		return nil, err
	}
	return &d, nil
}

func (d *DMI) ID() string {
	// ignore bios, because of bios can be rewritten via software
	tmp := d.Copy()
	tmp.BiosDate = ""
	tmp.BiosVendor = ""
	tmp.BiosVersion = ""
	return fmt.Sprintf("%x", sha256.Sum256(tmp.Bytes()))
}

func (d *DMI) Bytes() []byte {
	data, _ := json.Marshal(d)
	return data
}

func (d *DMI) String() string {
	var str string
	str += fmt.Sprintf("ID: %s\n", d.ID())
	str += fmt.Sprintf("\tBiosDate: %s\n", d.BiosDate)
	str += fmt.Sprintf("\tBiosVendor: %s\n", d.BiosVendor)
	str += fmt.Sprintf("\tBiosVersion: %s\n", d.BiosVersion)
	str += fmt.Sprintf("\tBoardName: %s\n", d.BoardName)
	str += fmt.Sprintf("\tBoardVendor: %s\n", d.BoardVendor)
	str += fmt.Sprintf("\tBoardVersion: %s\n", d.BoardVersion)
	str += fmt.Sprintf("\tBoardSerial: %s\n", d.BoardSerial)
	str += fmt.Sprintf("\tBoardAssetTag: %s\n", d.BoardAssetTag)
	str += fmt.Sprintf("\tProductName: %s\n", d.ProductName)
	str += fmt.Sprintf("\tProductFamily: %s\n", d.ProductFamily)
	str += fmt.Sprintf("\tProductVersion: %s\n", d.ProductVersion)
	str += fmt.Sprintf("\tProductSerial: %s\n", d.ProductSerial)
	str += fmt.Sprintf("\tProductUUID: %s\n", d.ProductUUID)
	return str
}

func (d *DMI) Copy() *DMI {
	return &DMI{
		BiosDate:       d.BiosDate,
		BiosVendor:     d.BiosVendor,
		BiosVersion:    d.BiosVersion,
		BoardName:      d.BoardName,
		BoardVendor:    d.BoardVendor,
		BoardVersion:   d.BoardVersion,
		BoardSerial:    d.BoardSerial,
		BoardAssetTag:  d.BoardAssetTag,
		ProductName:    d.ProductName,
		ProductFamily:  d.ProductFamily,
		ProductVersion: d.ProductVersion,
		ProductSerial:  d.ProductSerial,
		ProductUUID:    d.ProductUUID,
	}
}

func readFile(filename string) (string, error) {
	data, err := ioutil.ReadFile(filename)
	if err != nil {
		return "", err
	}
	l := len(data)
	if data[l-1] == '\n' {
		return string(data[:l-1]), nil
	}
	return string(data), nil
}
