package background

import (
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"pkg.deepin.io/daemon/sync/infrastructure/storage"
	"pkg.deepin.io/daemon/sync/infrastructure/utils"
	. "pkg.deepin.io/daemon/sync/modules/model"
)

type File struct {
	URI       string `json:"uri"`
	CloudName string `json:"cloud_name"`
	Md5       string `json:"md5"`
}
type FileList []*File

type Data struct {
	Version   string   `json:"version"`
	Workspace FileList `json:"workspace"`
	Greeter   *File    `json:"greeter"`
}

const (
	Name = "background"

	backgroundDir    = "Pictures/Wallpapers"
	cloudNamePrefix  = "workspace"
	backgroundMaxNum = 4
)

func NewData(version string, uris []string, greeter string) *Data {
	var files FileList
	for idx, uri := range uris {
		if idx >= backgroundMaxNum {
			break
		}
		md5, _ := utils.SumFileMd5(utils.TrimFileURI(uri))
		if len(md5) == 0 {
			md5 = filepath.Base(uri)
		}
		files = append(files, &File{
			URI:       uri,
			CloudName: fmt.Sprintf("%s%d", cloudNamePrefix, idx),
			Md5:       md5,
		})
	}
	var info = Data{
		Version:   version,
		Workspace: files,
	}
	if len(greeter) == 0 {
		info.Greeter = &File{}
		return &info
	}
	md5, _ := utils.SumFileMd5(utils.TrimFileURI(greeter))
	if len(md5) == 0 {
		md5 = filepath.Base(greeter)
	}
	info.Greeter = &File{
		URI:       greeter,
		CloudName: "greeter",
		Md5:       md5,
	}
	return &info
}

func (core *Data) GenCache(current DataIFC) DataIFC {
	curInfo := current.(*Data)
	length := len(curInfo.Workspace)
	var info = Data{
		Version:   core.Version,
		Workspace: core.Workspace[:length],
		Greeter:   core.Greeter,
	}
	return &info
}

func (core *Data) Merge(current, cache DataIFC) {
	curInfo := current.(*Data)
	cacheInfo := cache.(*Data)
	if !curInfo.Greeter.Equal(cacheInfo.Greeter) {
		core.Greeter = curInfo.Greeter
	} else {
		core.Greeter.URI = curInfo.Greeter.URI
	}

	curLen := len(curInfo.Workspace)
	cacheLen := len(cacheInfo.Workspace)
	rLen := len(core.Workspace)
	var idx = 0
	for _, v := range curInfo.Workspace {
		if idx >= cacheLen {
			break
		}
		if idx >= rLen {
			break
		}
		if !v.Equal(cacheInfo.Workspace[idx]) {
			core.Workspace[idx] = v
		} else {
			// fix uri with local uri
			core.Workspace.ChangeURI(core.Workspace[idx].URI, v.URI)
		}
		idx++
	}
	for ; idx < curLen; idx++ {
		if idx < rLen {
			core.Workspace[idx] = curInfo.Workspace[idx]
		} else {
			core.Workspace = append(core.Workspace, curInfo.Workspace[idx])
		}
	}
}

func (core *Data) Name() string {
	return Name
}

func (core *Data) Model() interface{} {
	return &Data{}
}

func (core *Data) ToIFC(v interface{}) DataIFC {
	return v.(*Data)
}

func (core *Data) Equal(target DataIFC) bool {
	info := target.(*Data)
	if !core.Workspace.Equal(info.Workspace) {
		return false
	}
	return core.Greeter.Equal(info.Greeter)
}

func (core *Data) Upload(cloud *utils.CloudStorage) error {
	for _, bg := range core.Workspace {
		err := core.uploadFile(cloud, bg)
		if err != nil {
			return err
		}
	}
	if core.Greeter == nil {
		return nil
	}
	return core.uploadFile(cloud, core.Greeter)
}

func (core *Data) Download(cloud *utils.CloudStorage) error {
	for _, bg := range core.Workspace {
		err := core.downloadFile(cloud, bg)
		if err != nil {
			return err
		}
	}
	if core.Greeter == nil {
		return nil
	}
	return core.downloadFile(cloud, core.Greeter)
}

func (core *Data) uploadFile(cloud *utils.CloudStorage, info *File) error {
	prefix := utils.GetFilePath(Name, info.CloudName)
	// limit file size < 5M
	filename := utils.TrimFileURI(info.URI)
	if err := checkFile(filename); err != nil {
		return err
	}
	data, err := ioutil.ReadFile(filename)
	if err != nil {
		return err
	}
	_, err = cloud.Set(prefix, data, &storage.SetOptions{NewVersion: true})
	return err
}

func (core *Data) downloadFile(cloud *utils.CloudStorage, info *File) error {
	if utils.IsFileExist(utils.TrimFileURI(info.URI)) {
		return nil
	}
	prefix := utils.GetFilePath(Name, info.CloudName)
	ret, err := cloud.Get(prefix,
		&storage.GetOptions{Version: storage.VersionLatest})
	if err != nil {
		return err
	}
	uri, err := writeBackground(ret.Content, info.URI)
	if err != nil {
		return err
	}
	info.URI = uri
	return nil
}

func (list FileList) Equal(infos FileList) bool {
	len1 := len(list)
	len2 := len(infos)
	if len1 != len2 {
		return false
	}
	for i := 0; i < len1; i++ {
		if !list[i].Equal(infos[i]) {
			return false
		}
	}
	return true
}

func (list FileList) Files() []string {
	var ret []string
	for _, v := range list {
		ret = append(ret, v.URI)
	}
	return ret
}

func (list FileList) ChangeURI(src, dst string) {
	for _, v := range list {
		if v.URI == src {
			v.URI = dst
		}
	}
}

func (d *File) Equal(info *File) bool {
	return d.CloudName == info.CloudName && d.Md5 == info.Md5
}

func writeBackground(data []byte, uri string) (string, error) {
	uri = utils.TrimFileURI(uri)
	var perr *os.PathError
	err := os.MkdirAll(filepath.Dir(uri), 0755)
	if err == nil {
		err = ioutil.WriteFile(uri, data, 0644)
		if err != nil {
			perr = err.(*os.PathError)
			if perr.Err.Error() == os.ErrPermission.Error() {
				goto create
			}
			return "", err
		} else {
			return "file://" + uri, nil
		}
	}

	perr = err.(*os.PathError)
	if perr.Err.Error() != os.ErrPermission.Error() {
		return "", err
	}
create:
	dir := filepath.Join(os.Getenv("HOME"), backgroundDir)
	err = os.MkdirAll(dir, 0755)
	if err != nil {
		return "", err
	}
	newfile := filepath.Join(dir, filepath.Base(uri))
	return "file://" + newfile, ioutil.WriteFile(newfile, data, 0755)
}

func checkFile(filename string) error {
	info, err := os.Stat(filename)
	if err != nil {
		return err
	}
	// limit file size < 5M
	if info.Size() > 5*1024*1024 {
		return fmt.Errorf("file is too large: %s", filename)
	}
	return nil
}
