package modules

import (
	"encoding/json"
	Logger "pkg.deepin.io/daemon/sync/infrastructure/log"
	"pkg.deepin.io/daemon/sync/infrastructure/storage"
	"pkg.deepin.io/daemon/sync/infrastructure/utils"
	"pkg.deepin.io/daemon/sync/modules/appearance"
	"pkg.deepin.io/daemon/sync/modules/audio"
	"pkg.deepin.io/daemon/sync/modules/background"
	"pkg.deepin.io/daemon/sync/modules/dock"
	"pkg.deepin.io/daemon/sync/modules/launcher"
	. "pkg.deepin.io/daemon/sync/modules/model"
	"pkg.deepin.io/daemon/sync/modules/network"
	"pkg.deepin.io/daemon/sync/modules/peripherals"
	"pkg.deepin.io/daemon/sync/modules/screenedge"
	"pkg.deepin.io/daemon/sync/modules/updater"
	"sync"
	"time"
)

type Manager struct {
	cloudReset bool
	hwID       string
	userID     *int64
	token      storage.TokenSource
	switcher   *Switcher

	itemBusSet    map[string]*ModuleBusInfo
	itemBusLocker sync.RWMutex

	items       DataIFCList
	itemsLocker sync.RWMutex
	itemDataMap map[string]DataIFC
}

const (
	cloudDuration = time.Second * 5
)

func NewManager(userID *int64, hwID string, token storage.TokenSource,
	reset bool) (*Manager, error) {
	var m = Manager{
		cloudReset:  reset,
		hwID:        hwID,
		userID:      userID,
		token:       token,
		items:       DataIFCList{},
		itemBusSet:  make(map[string]*ModuleBusInfo),
		itemDataMap: make(map[string]DataIFC),
	}
	m.initItemDataMap()
	m.loadSwitcher()

	return &m, nil
}

func (m *Manager) Register(module string, info *ModuleBusInfo) error {
	m.setItem(module)
	m.setItemBus(module, info)
	if module == appearance.Name {
		m.setItem(background.Name)
		m.setItemBus(background.Name, info)
	}
	return nil
}

func (m *Manager) SwitcherSet(name string, enabled bool) error {
	err := m.switcher.Enable(name, enabled)
	if err != nil {
		return err
	}
	return m.switcher.WriteFile(localSwitcherFile)
}

func (m *Manager) SwitcherGet(name string) bool {
	return m.switcher.IsEnabled(name)
}

func (m *Manager) SwitcherDump() ([]byte, error) {
	return m.switcher.Bytes()
}

func (m *Manager) getCloudStorage() (*utils.CloudStorage, error) {
	cloud, err := utils.CloudConnect(m.token, m.hwID)
	if err != nil {
		Logger.Warning("[Modules] Failed to connect cloud:", err)
		return nil, err
	}
	return cloud, nil
}

func (m *Manager) loadSwitcher() {
	s, err := newSwitcher(ConfigSwitcherFile)
	if err != nil {
		Logger.Warning("[Modules] Failed to load system switcher:", err)
	}
	m.switcher = s

	s, err = newSwitcher(localSwitcherFile)
	if err != nil {
		return
	}
	m.switcher.Fixed(s)
}

func (m *Manager) initItemDataMap() {
	m.itemDataMap[appearance.Name] = &appearance.Data{}
	m.itemDataMap[background.Name] = &background.Data{}
	m.itemDataMap[dock.Name] = &dock.Data{}
	m.itemDataMap[launcher.Name] = &launcher.Data{}
	m.itemDataMap[network.Name] = &network.Data{}
	m.itemDataMap[peripherals.Name] = &peripherals.Data{}
	m.itemDataMap[screenedge.Name] = &screenedge.Data{}
	m.itemDataMap[audio.Name] = &audio.Data{}
	m.itemDataMap[updater.Name] = &updater.Data{}
}

func (m *Manager) getModuleData(name string) ([]byte, error) {
	if !m.SwitcherGet(name) {
		Logger.Debug("[Modules] the module has been disabled:",
			name)
		return nil, nil
	}
	busInfo := m.getItemBus(name)
	if busInfo == nil {
		Logger.Debug("[Modules] the module no bus info register:",
			name)
		return nil, nil
	}
	data, err := busInfo.Get()
	if err != nil {
		return nil, err
	}
	// Logger.Info("[Module] Get data:", name, string(data))
	if name != appearance.Name && name != background.Name {
		return data, nil
	}
	var wrapper appearanceWrapper
	err = json.Unmarshal(data, &wrapper)
	if err != nil {
		return nil, err
	}
	if name == appearance.Name {
		return json.Marshal(&wrapper.Data)
	}
	bgInfo := background.NewData(wrapper.Version,
		wrapper.BackgroundURIs, wrapper.GreeterBackground)
	return json.Marshal(bgInfo)
}

func (m *Manager) setModuleData(name string, data []byte) error {
	busInfo := m.getItemBus(name)
	if busInfo == nil {
		Logger.Debug("[Modules] the module no bus info register:",
			name)
		return nil
	}
	if name != appearance.Name && name != background.Name {
		// Logger.Info("[Modules] [Set] local data:", name, string(data))
		return busInfo.Set(data)
	}
	locData, err := busInfo.Get()
	if err != nil {
		return err
	}
	var wrapper appearanceWrapper
	err = json.Unmarshal(locData, &wrapper)
	if err != nil {
		return err
	}
	if name == appearance.Name {
		err = wrapper.FixAppearance(data)
	} else {
		err = wrapper.FixBackground(data)
	}
	if err != nil {
		return err
	}
	// Logger.Info("[Modules] [Set] local data:", name, string(wrapper.Bytes()))
	return busInfo.Set(wrapper.Bytes())
}

func (m *Manager) setItemBus(name string, info *ModuleBusInfo) {
	m.itemBusLocker.Lock()
	m.itemBusSet[name] = info
	m.itemBusLocker.Unlock()
}

func (m *Manager) getItemBus(name string) *ModuleBusInfo {
	m.itemBusLocker.RLock()
	defer m.itemBusLocker.RUnlock()
	return m.itemBusSet[name]
}

func (m *Manager) setItem(name string) {
	m.itemsLocker.Lock()
	defer m.itemsLocker.Unlock()
	info, ok := m.itemDataMap[name]
	if !ok {
		Logger.Warning("Unspported module:", name)
		return
	}
	m.items = m.items.Append(name, info)
}

func (m *Manager) getItem(name string) DataIFC {
	m.itemsLocker.RLock()
	defer m.itemsLocker.RUnlock()
	return m.items.Get(name)
}
