package modules

import (
	"encoding/json"
	Logger "pkg.deepin.io/daemon/sync/infrastructure/log"
	"pkg.deepin.io/daemon/sync/infrastructure/storage"
	"pkg.deepin.io/daemon/sync/infrastructure/utils"
	"pkg.deepin.io/daemon/sync/modules/background"
	. "pkg.deepin.io/daemon/sync/modules/model"
)

type ModuleIndex struct {
	Appearance  string `json:"appearance"`
	Audio       string `json:"audio"`
	Background  string `json:"background"`
	Dock        string `json:"dock"`
	Launcher    string `json:"launcher"`
	Network     string `json:"network"`
	Peripherals string `json:"peripherals"`
	ScreenEdge  string `json:"screen_edge"`
	Updater     string `json:"updater"`
}

type SyncError struct {
	Code    StateType `json:"code"`
	Module  string    `json:"module"`
	Message string    `json:"message"`
}

type StateType int32

const (
	cloudKeyIndex = "index"
)

const (
	CodeSynchronizing StateType = 100
	CodeSuccess       StateType = 200

	CodeLoginErr StateType = iota + 1000
	CodeNetwrokErr
	CodeTokenErr
	CodeIndexUploadErr
	CodeIndexDownloadErr
	CodeIndexFormatErr
	CodeModuleUploadErr
	CodeModuleDownloadErr
	CodeModuleSetErr
	CodeModuleGetErr
	CodeModuleFormatErr
)

func (code StateType) Message() string {
	switch code {
	case CodeSynchronizing:
		return "Synchronizing"
	case CodeSuccess:
		return "Synchronized"
	}
	return ""
}

func (e SyncError) Error() string {
	data, _ := json.Marshal(&e)
	return string(data)
}

func NewSyncError(code StateType, module, msg string) *SyncError {
	return &SyncError{
		Code:    code,
		Module:  module,
		Message: msg,
	}
}

func (m *Manager) Sync() (bool, *SyncError) {
	Logger.Debug("[Modules] start sync...")
	if !m.SwitcherGet(switcherEnabled) {
		Logger.Debug("[Modules] sync has been disabled")
		return false, nil
	}
	cloud, err := m.getCloudStorage()
	if cloud == nil {
		return false, NewSyncError(CodeTokenErr, "", err.Error())
	}
	idxData, err := cloud.GetLatest(cloudKeyIndex)
	if err != nil {
		if err != storage.ErrNotFound {
			return false, NewSyncError(CodeIndexDownloadErr,
				cloudKeyIndex, err.Error())
		}
		idxData = &storage.GetResult{}
	}
	var idxInfo ModuleIndex
	if len(idxData.Content) != 0 {
		err = json.Unmarshal(idxData.Content, &idxInfo)
		if err != nil {
			return false, NewSyncError(CodeIndexFormatErr,
				cloudKeyIndex, err.Error())
		}
	}
	var idxSet = make(map[string]string)
	for _, item := range m.items {
		if !m.SwitcherGet(item.Name()) {
			Logger.Debugf("The module(%s) has been disabled",
				item.Name())
			continue
		}
		var itemVersion string
		var e *SyncError
		if m.cloudReset {
			itemVersion, e = m.ResetModule(cloud, item)
		} else {
			itemVersion, e = m.SyncModule(cloud, item,
				idxInfo.Get(item.Name()))
		}
		if e != nil {
			Logger.Warning("[Modules] [Sync] Failed:", item.Name(), e)
			continue
		}
		if len(itemVersion) != 0 {
			idxSet[item.Name()] = itemVersion
		}
	}
	if len(idxSet) == 0 {
		Logger.Debug("[Modules] no changes")
		return false, nil
	}

	// update index
	for k, v := range idxSet {
		idxInfo.Update(k, v)
	}
	Logger.Debug("[Modules] will update index to:", string(idxInfo.Bytes()))
	_, err = cloud.Set(cloudKeyIndex, idxInfo.Bytes(),
		&storage.SetOptions{NewVersion: true})
	if err != nil {
		return false, NewSyncError(CodeIndexUploadErr,
			cloudKeyIndex, err.Error())
	}
	return true, nil
}

func (m *Manager) ResetModule(cloud *utils.CloudStorage,
	item DataIFC) (string, *SyncError) {
	var name = item.Name()
	data, err := m.getModuleData(name)
	if err != nil {
		return "", NewSyncError(CodeModuleGetErr, name, err.Error())
	}
	// background need upload files firstly
	if name == background.Name {
		var info background.Data
		err = json.Unmarshal(data, &info)
		if err != nil {
			return "", NewSyncError(CodeModuleFormatErr, name, err.Error())
		}
		err = info.Upload(cloud)
		if err != nil {
			return "", NewSyncError(CodeModuleUploadErr, name, err.Error())
		}
	}
	version, err := cloud.Set(name, data,
		&storage.SetOptions{NewVersion: true})
	if err != nil {
		return "", NewSyncError(CodeModuleUploadErr, name, err.Error())
	}
	// save cache
	utils.SaveLocalCache(name, *m.userID, &utils.LocalCache{
		Data:    data,
		Version: version,
	})
	return version, nil
}

func (m *Manager) SyncModule(cloud *utils.CloudStorage, item DataIFC,
	version string) (string, *SyncError) {
	var upData []byte
	var cacheData []byte
	var cacheRet *utils.LocalCache
	var name = item.Name()
	var rInfo = item.Model()
	var cacheInfo = item.Model()
	var rIFC, curIFC, cacheIFC DataIFC

	// got current data
	data, err := m.getModuleData(name)
	if err != nil {
		return "", NewSyncError(CodeModuleGetErr, name, err.Error())
	}
	if len(data) == 0 {
		// disabled or no register
		return "", nil
	}
	var curInfo = item.Model()
	err = json.Unmarshal(data, curInfo)
	if err != nil {
		return "", NewSyncError(CodeModuleFormatErr, name, err.Error())
	}
	curIFC = item.ToIFC(curInfo)

	// got cloud data
	ret, err := cloud.Get(name, &storage.GetOptions{Version: version})
	if err != nil {
		if err != storage.ErrNotFound {
			return "", NewSyncError(CodeModuleDownloadErr,
				name, err.Error())
		}
		ret = &storage.GetResult{}
	}
	if len(ret.Content) == 0 {
		// never uploaded, uploading now
		upData = data
		cacheData = data
		goto upload
	}
	err = json.Unmarshal(ret.Content, rInfo)
	if err != nil {
		return "", NewSyncError(CodeModuleFormatErr, name, err.Error())
	}
	rIFC = item.ToIFC(rInfo)

	// got cache data
	cacheRet, _ = utils.LoadLocalCache(name, *m.userID)
	if cacheRet == nil || len(cacheRet.Data) == 0 {
		// never synchronized, synchronizing now and save cache
		err = m.handleBackgroundFiles(name, rIFC, false)
		if err != nil {
			return "", NewSyncError(CodeModuleDownloadErr, name, err.Error())
		}
		return "", m.setModuleAndSaveCache(name,
			jsonMarshal(rIFC.GenCache(curIFC)), ret.Version)
	}
	// Logger.Info("[Modules] [Sync] cache data:", name, string(cacheRet.Data))
	err = json.Unmarshal(cacheRet.Data, cacheInfo)
	if err != nil {
		// local cache invalid, replace it with cloud data
		err = m.handleBackgroundFiles(name, rIFC, false)
		if err != nil {
			return "", NewSyncError(CodeModuleDownloadErr, name, err.Error())
		}
		return "", m.setModuleAndSaveCache(name,
			jsonMarshal(rIFC.GenCache(curIFC)), ret.Version)
	}
	cacheIFC = item.ToIFC(cacheInfo)

	// compare remote, current and cache data
	if curIFC.Equal(cacheIFC) {
		// the local data no changes
		if ret.Version > cacheRet.Version {
			return ret.Version, m.setModuleAndSaveCache(name,
				jsonMarshal(rIFC.GenCache(curIFC)), ret.Version)
		}
		return "", nil
	}

	// the local data has changes
	// firstly merge the no changes options, then upload data
	rIFC.Merge(curIFC, cacheIFC)
	// upload backfround files
	err = m.handleBackgroundFiles(name, rIFC, true)
	if err != nil {
		return "", NewSyncError(CodeModuleDownloadErr, name, err.Error())
	}
	upData = jsonMarshal(rIFC)
	cacheData = jsonMarshal(rIFC.GenCache(curIFC))
upload:
	ver, err := cloud.Set(name, upData, &storage.SetOptions{NewVersion: true})
	if err != nil {
		return "", NewSyncError(CodeModuleUploadErr, name, err.Error())
	}
	m.setModuleAndSaveCache(name, cacheData, ver)
	return ver, nil
}

func (m *Manager) handleBackgroundFiles(name string, ifc DataIFC, upload bool) error {
	if name != background.Name {
		return nil
	}
	cloud, err := m.getCloudStorage()
	if err != nil {
		return err
	}
	info := ifc.(*background.Data)
	err = info.Download(cloud)
	if err != nil {
		return err
	}
	if !upload {
		return nil
	}
	return info.Upload(cloud)
}

func (m *Manager) setModuleAndSaveCache(name string, data []byte,
	version string) *SyncError {
	err := m.setModuleData(name, data)
	if err != nil {
		return NewSyncError(CodeModuleSetErr,
			name, err.Error())
	}
	utils.SaveLocalCache(name, *m.userID, &utils.LocalCache{
		Data:    data,
		Version: version,
	})
	return nil
}

func (index *ModuleIndex) Update(name, version string) bool {
	var changed = false
	switch name {
	case "appearance":
		changed = updateStrValue(&index.Appearance, version)
	case "audio":
		changed = updateStrValue(&index.Audio, version)
	case "background":
		changed = updateStrValue(&index.Background, version)
	case "dock":
		changed = updateStrValue(&index.Dock, version)
	case "launcher":
		changed = updateStrValue(&index.Launcher, version)
	case "network":
		changed = updateStrValue(&index.Network, version)
	case "peripherals":
		changed = updateStrValue(&index.Peripherals, version)
	case "screenedge":
		changed = updateStrValue(&index.ScreenEdge, version)
	case "updater":
		changed = updateStrValue(&index.Updater, version)
	}
	return changed
}

func (index *ModuleIndex) Get(name string) string {
	switch name {
	case "appearance":
		return index.Appearance
	case "audio":
		return index.Audio
	case "background":
		return index.Background
	case "dock":
		return index.Dock
	case "launcher":
		return index.Launcher
	case "network":
		return index.Network
	case "peripherals":
		return index.Peripherals
	case "screenedge":
		return index.ScreenEdge
	case "updater":
		return index.Updater
	}
	return ""
}

func (index *ModuleIndex) Bytes() []byte {
	data, _ := json.Marshal(index)
	return data
}

func updateStrValue(src *string, value string) bool {
	if *src == value {
		return false
	}
	*src = value
	return true
}

func jsonMarshal(v interface{}) []byte {
	data, _ := json.Marshal(v)
	return data
}
